package main

import (
	"log"
	"net"
	"github.com/songgao/water"
	"github.com/songgao/water/waterutil"
	"github.com/songgao/packets/ethernet"
	"github.com/google/gopacket"
	"github.com/google/gopacket/layers"
	"time"
	"runtime"
	"fmt"
	_ "net/http/pprof"
	"net/http"
	"io"
)

/*************************************************************************************************************************

for windows
	netsh interface ip set address name = "tap" source= static addr = 11.1.0.1 mask = 255.255.255.0 gateway = none
	connection mysql gui client ,sample:11.1.0.11:3306

go tool pprof --seconds 30 http://localhost:6060/debug/pprof/profile

*************************************************************************************************************************/

var bindServer = "127.0.0.1:3306" // mysql server

const SERVER_IP = "11.1.0.1"
const SERVER_PORT layers.TCPPort = 8000

type Port int
type TergetAddr struct {
	ip     net.IP
	port   Port
	delete bool
}

var mapping map[Port]*TergetAddr = make(map[Port]*TergetAddr)

func init() {
	runtime.GOMAXPROCS(runtime.NumCPU())
	log.SetFlags(log.Lshortfile | log.LstdFlags)
	//pprof web server
	go http.ListenAndServe("localhost:6060", nil)
}

func main() {
	if nic, err := water.New(water.Config{DeviceType: water.TAP}); err == nil {
		log.Printf("network interface card name is %s\r\n", nic.Name())
		time.Sleep(5 * time.Second)
		go tcpPortForwardServer(fmt.Sprintf("%s:%d", SERVER_IP, SERVER_PORT))
		NICRWLoop(nic)
	} else {
		log.Fatal(err)
	}
}

func NICRWLoop(nic *water.Interface) {
	for {
		var frame ethernet.Frame
		frame.Resize(1500)
		n, err := nic.Read([]byte(frame))
		if err != nil {
			log.Panicln(err)
		}
		go func(frame ethernet.Frame) {
			//arp
			if frame.Ethertype() == ethernet.ARP {
				arpRsp(nic, frame)
				return
			}
			//ipv4
			if frame.Ethertype() == ethernet.IPv4 && waterutil.IPv4Protocol(frame.Payload()) == waterutil.TCP {
				tcpRsp(nic, frame)
				return
			}
		}(frame[:n])
	}
}

func tcpRsp(nic *water.Interface, frame ethernet.Frame) {
	var eth layers.Ethernet
	var ipv4 layers.IPv4
	var tcp layers.TCP
	var payload gopacket.Payload
	parser := gopacket.NewDecodingLayerParser(layers.LayerTypeEthernet, &eth, &ipv4, &tcp, &payload)
	err := parser.DecodeLayers(frame, &[]gopacket.LayerType{})
	if err != nil {
		log.Panicln(err)
	}
	if isFromClient(&ipv4, &tcp) {
		if tcp.SYN {
			mapping[Port(tcp.SrcPort)] = &TergetAddr{ip: ipv4.DstIP, port: Port(tcp.DstPort)}
		}
		if tcp.ACK {
			if entry := mapping[Port(tcp.SrcPort)]; entry != nil && entry.delete {
				delete(mapping, Port(tcp.SrcPort))
				log.Println("map size is ", len(mapping))
			}
		}
		tcp.DstPort = SERVER_PORT
	} else {
		if entry := mapping[Port(tcp.DstPort)]; entry != nil {
			if tcp.FIN && tcp.ACK {
				mapping[Port(tcp.DstPort)].delete = true
			}
			tcp.SrcPort = layers.TCPPort(entry.port)
		}
	}
	tcp.SetNetworkLayerForChecksum(&ipv4)
	ethReverse(&eth)
	ipv4Reverse(&ipv4)
	buf := gopacket.NewSerializeBuffer()
	opts := gopacket.SerializeOptions{FixLengths: true, ComputeChecksums: true}
	err = gopacket.SerializeLayers(buf, opts, &eth, &ipv4, &tcp, &payload)
	if err != nil {
		log.Panicln(err)
	}
	go nic.Write(buf.Bytes())
}

func isFromClient(ipv4 *layers.IPv4, tcp *layers.TCP) bool {
	if ipv4.SrcIP.String() == SERVER_IP && tcp.SrcPort == SERVER_PORT {
		return false
	}
	return true
}

func arpRsp(nic *water.Interface, frame ethernet.Frame) {
	var eth layers.Ethernet
	var arp layers.ARP
	parser := gopacket.NewDecodingLayerParser(layers.LayerTypeEthernet, &eth, &arp)
	err := parser.DecodeLayers(frame, &[]gopacket.LayerType{})
	if err != nil {
		log.Panicln(err)
	}
	ethReverse(&eth)
	arp.Operation = layers.ARPReply
	arpSrcMac := arp.SourceHwAddress
	arpSrcIp := arp.SourceProtAddress
	arpDstIp := arp.DstProtAddress
	arp.SourceProtAddress = arpDstIp
	arp.DstHwAddress = arpSrcMac
	arp.DstProtAddress = arpSrcIp
	buf := gopacket.NewSerializeBuffer()
	opts := gopacket.SerializeOptions{FixLengths: true, ComputeChecksums: true}
	err = gopacket.SerializeLayers(buf, opts, &eth, &arp)
	if err != nil {
		log.Panicln(err)
	}
	nic.Write(buf.Bytes())
}

func ethReverse(eth *layers.Ethernet) {
	srcMac := eth.SrcMAC
	dstMac := eth.DstMAC
	eth.SrcMAC = dstMac
	eth.DstMAC = srcMac
}

func ipv4Reverse(ipv4 *layers.IPv4) {
	dstIp := ipv4.DstIP
	srcIp := ipv4.SrcIP
	ipv4.DstIP = srcIp
	ipv4.SrcIP = dstIp
}

func tcpPortForwardServer(adderss string) {
	lis, err := net.Listen("tcp", adderss)
	if err != nil {
		log.Println(err)
		return
	}
	log.Println("tcp server ")
	defer lis.Close()
	for {
		conn, err := lis.Accept()
		if err != nil {
			continue
		}
		go func(sconn net.Conn) {
			ExitChan := make(chan bool)
			dconn, err := net.Dial("tcp", bindServer)
			defer dconn.Close()
			defer sconn.Close()
			if err != nil {
				return
			}
			go func(sconn net.Conn, dconn net.Conn, Exit chan bool) {
				io.Copy(dconn, sconn)
				ExitChan <- true
			}(sconn, dconn, ExitChan)
			go func(sconn net.Conn, dconn net.Conn, Exit chan bool) {
				io.Copy(sconn, dconn)
				ExitChan <- true
			}(sconn, dconn, ExitChan)
			<-ExitChan
		}(conn)
	}
}
